/* -*- c-mode -*- */

%module "XML::JHXML"
%{
#include "b4302lib.h"
%}

%pragma(perl5) include="JHXML_include.pl";

%inline %{
  int callback_err(char *sz)
    {
      croak("JHXLM error: %s\n ",sz);
      return 0;
    } /* of callback_err */ 
  int callback_msg(char *sz)
    {
      /* supposed to return and continue */
      warn("JHXML warning: %s\n ",sz);
      return 0;
    } /* of callback_msg */
%};

/* get rid of the b4302 prefix */
%rename b4302DocGetRootElement  DocGetRootElement;
%rename b4302GetNextElement     GetNextElement;
%rename b4302GetPrevElement     GetPrevElement;
%rename b4302GetFirstChild      GetFirstChild;
%rename b4302GetLastChild       GetLastChild;
%rename b4302GetParent          GetParent;
%rename b4302GetFirstAttr       GetFirstAttr;
%rename b4302GetNextAttr        GetNextAttr;
%rename b4302GetAttribute       GetAttribute;
%rename b4302GetAttributeNode   GetAttributeNode;
%rename b4302SetAttribute       SetAttribute;
%rename b4302GetData            GetData;
%rename b4302ParseFile          ParseFile;
%rename b4302WriteFile          WriteFile;
%rename b4302ParseString        ParseString;
%rename b4302GetElementName     GetElementName;

%rename b4302InsertBefore       InsertBefore;
%rename b4302InsertAfter        InsertAfter;
%rename b4302RemoveChild        RemoveChild;
%rename b4302ReplaceChild       ReplaceChild;
%rename b4302CreateElement      CreateElement;
%rename b4302CreateTextNode     CreateTextNode;
%rename b4302DestroyNode        DestroyNode;
%rename b4302DestroySubtree     DestroySubtree;
%rename b4302DestroyDoc         DestroyDoc;
%rename b4302DocSetAttribute    DocSetAttribute;
%rename b4302DocRemoveAttribute DocRemoveAttribute;
%rename b4302RemoveAttribute    RemoveAttribute;
%rename b4302SetData            SetData;
%rename b4302DocAppendText      DocAppendText;
%rename b4302DocAppendTextChild DocAppendTextChild;
%rename b4302AppendText         AppendText;
%rename b4302AppendTextChild    AppendTextChild;

%rename b4302NewOptions         NewParserOptions;
%rename b4302DeleteOptions      DeleteParserOptions;
%rename b4302SetOptionParseErrorsOK SetOptionParseErrorsOK;
%rename b4302SetOptionNoBlanks  SetOptionNoBlanks;
%rename b4302SetOptionIndent    SetOptionIndent;

%rename b4302ToString           ToString;

%rename b4302FindFirstChild     FindFirstChild;
%rename b4302FindNextSibling    FindNextSibling;
%rename b4302FindPrevSibling    FindPrevSibling;
%rename b4302FindAncestor       FindAncestor;



/* set utf8 flag for char * returned from functions */
%typemap(perl5,out) char * {
  $result = sv_newmortal();
  sv_setpv((SV *)$result, (char *) ($1));
  SvUTF8_on((SV*)($result));
  argvi++;
};

%name(Attr) typedef struct b4302Attr {
  char *pszName;
  char *pszValue;
} b4302AttrType;

%name(ParserOptions) typedef struct b4302IOOptions {
  char szCode[200];
  int fParseErrorsOK;
  int fNoBlanks;
  int iIndent;
} b4302IOOptionsType;


%inline %{
  int IsSameNode( b4302ElemType* self, b4302ElemType* node ) { return ((self == node) ? 1 : 0); }
  int HasChildNodes( b4302ElemType* self ) { return ((void *) b4302GetFirstChild(self) != NULL); }
  int IsTextNode( b4302ElemType* self ) { return (*(self->pszName) == '<'); }
  int AppendChild(b4302ElemType *peParent, 
		  b4302ElemType *peNew) { 
    return b4302InsertBefore(peParent,peNew, (b4302ElemType *) NULL);
  }
%}

b4302ElemType *b4302DocGetRootElement(b4302DocType *pDoc);
b4302ElemType *b4302GetNextElement(b4302ElemType *pElem);
b4302ElemType *b4302GetPrevElement(b4302ElemType *pElem);
b4302ElemType *b4302GetFirstChild(b4302ElemType *pElem);
b4302ElemType *b4302GetLastChild(b4302ElemType *pElem);
b4302ElemType *b4302GetParent(b4302ElemType *pElem);
b4302AttrType *b4302GetFirstAttr(b4302ElemType *pElem);
b4302AttrType *b4302GetNextAttr(b4302AttrType *pAttr);
char * b4302GetAttribute(b4302ElemType *pElem, const char *pszAttrName);
b4302AttrType * b4302GetAttributeNode(b4302ElemType *pElem, 
				      const char *pszAttrName);
b4302AttrType * b4302SetAttribute(b4302ElemType *pElem, 
					 const char *pszAttrName, 
					 const char *pszAttrValue);
char *b4302GetData(b4302ElemType *pElem);

int b4302WriteFile(b4302DocType *pDoc, char *pszFileName);
b4302DocType *b4302ParseFile(char *pszFileName, b4302IOOptionsType *po);
b4302DocType *b4302ParseString(char *pszStringXML, b4302IOOptionsType *po);

char *b4302GetElementName(b4302ElemType *pElem);
int b4302InsertBefore(b4302ElemType *peParent, 
		      b4302ElemType *peNew, 
		      b4302ElemType *peRef);

int b4302InsertAfter(b4302ElemType *peParent, 
		      b4302ElemType *peNew, 
		      b4302ElemType *peRef);

b4302ElemType *b4302RemoveChild(b4302ElemType *peParent, 
				b4302ElemType *peChildToRemove);

b4302ElemType *b4302ReplaceChild(b4302ElemType *peParent, 
				 b4302ElemType *peNew,
				 b4302ElemType *peOld);

b4302ElemType *b4302CreateElement(b4302DocType *pd, char *pszElemName);

b4302ElemType *b4302CreateTextNode(b4302DocType *pd, char *pszString);

int b4302DestroyNode(b4302ElemType *peNode);

int b4302DestroySubtree(b4302ElemType *peSub);

int b4302DestroyDoc(b4302DocType *pd);
b4302AttrType * b4302DocSetAttribute(b4302DocType *pd,
					    b4302ElemType *pElem, 
					    const char *pszAttrName, 
					    const char *pszAttrValue);
int b4302DocRemoveAttribute(b4302DocType *pd,
			    b4302ElemType *pElem, 
			    const char *pszAttrName);
int b4302RemoveAttribute(b4302ElemType *pElem, 
			 const char *pszAttrName);

char *b4302SetData(b4302ElemType *pElem, const char *pszString);
b4302ElemType *b4302DocAppendText(b4302DocType *pd,
				  b4302ElemType *peParent, 
				  const char *pszString);
b4302ElemType *b4302AppendText(b4302ElemType *pElemParent, 
			       const char *pszString);
b4302ElemType *b4302DocAppendTextChild(b4302DocType *pd, 
				       b4302ElemType *peParent, 
				       const char *pszElemName,
				       const char *pszString);
b4302ElemType *b4302AppendTextChild(b4302ElemType *pElemParent, 
				    const char *pszElemName,
				    const char *pszString);

b4302ElemType *b4302FindFirstChild(b4302ElemType *peParent, char *pszStr);
b4302ElemType *b4302FindNextSibling(b4302ElemType *peNode, char *pszStr);
b4302ElemType *b4302FindPrevSibling(b4302ElemType *peNode, char *pszStr);
b4302ElemType *b4302FindAncestor(b4302ElemType *peChild, char *pszStr);
b4302IOOptionsType *b4302NewOptions(void);

void b4302DeleteOptions(b4302IOOptionsType *);
void b4302SetOptionParseErrorsOK(b4302IOOptionsType *po, int fFlag);
void b4302SetOptionNoBlanks(b4302IOOptionsType *po, int fFlag);
void b4302SetOptionIndent(b4302IOOptionsType *po, int i);


%typemap(newfree) 	char * "free($1);";
%newobject b4302ToString; 
char *b4302ToString(b4302ElemType *pe);
