/* XML small lib header file b4302lib.h */
#ifndef B4302LIBH

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "b4302libdef.h"

#if defined WIN32 || defined NOINLINE
#define inline
#endif

extern char *pszb4302Text;
extern int cb4302Len;
extern b4302Events eb4302Token;
extern b4302fParsing;
extern char *b4302szComment;

#define XMLVER10 "1.0"
#define XMLVER11 "1.1"
#define XMLENC_UTF8 "utf-8"
#define XMLENC_UTF16 "utf-16"
#define XMLENC_88591 "iso-8859-1"
#define XMLENC_88592 "iso-8859-2"
#define XMLENC_88593 "iso-8859-3"
#define XMLENC_88594 "iso-8859-4"
#define XMLENC_88595 "iso-8859-5"
#define XMLENC_88596 "iso-8859-6"
#define XMLENC_88597 "iso-8859-7"
#define XMLENC_88598 "iso-8859-8"
#define XMLENC_88599 "iso-8859-9"

#define ADDSTRING 1
#define SEARCHONLY 0
#define TEXTPSEUDOTAGNAME "<"
#define ROOTPSEUDOTAGNAME ">"
#define PSEUDOOTHER '/'
#define COMMENTPSEUDOTAGNAME "/COMMENT/"
#define CDATAPSEUDOTAGNAME "/CDATA/"
#define PIPSEUDOTAGNAME "/PI/"


#define b4302peNULL (b4302ElemType *) NULL
#define b4302paNULL (b4302AttrType *) NULL
#define b4302pdNULL (b4302DocType *) NULL
#define b4302pstrNULL (b4302StringType *) NULL
#define b4302pslNULL (b4302StringListType *) NULL
#define b4302psstrNULL (b4302StringSType *) NULL
#define b4302psslNULL (b4302StringSListType *) NULL
#define b4302psaNULL (b4302StringAreaType *) NULL
#define b4302psiNULL (b4302StackItemType *) NULL
#define b4302poNULL (b4302IOOptionsType *) NULL
#define b4302cNULL (char *) NULL

#define b4302HASHMAX 128
#define b4302HASHMASK 127
#define b4302SHASHMAX 128
#define b4302SHASHMASK 127

typedef struct b4302String { /* one of strings, linked in a list */
  char *szString; /* allocated in global area, dynamically */
  struct b4302String *pNext; /* next string in a set */
} b4302StringType;

typedef struct b4302StringList { /* string set (implemented as a list) anchor */
  int cNumberOfStrings; /* number of strings */
  struct b4302String *pFirst[b4302HASHMAX]; /* first string (main 
					       entry point to set) */
} b4302StringListType;

typedef struct b4302StringS { /* one of strings, linked in a list */
  char *szString; /* allocated in global area, dynamically */
  struct b4302StringList *pslSecondary; /* secondary lst, such as for values */
  struct b4302StringS *psstrNext; /* next string in a set */
} b4302StringSType;

typedef struct b4302StringSList { /* string set (implemented as a list) anchor */
  int csstrNumberOfStrings; /* number of strings */
  struct b4302StringS *psstrFirst[b4302SHASHMAX]; /* first string, trivially 
						     hashed (main entry 
						     point to set) */
} b4302StringSListType;

typedef struct b4302StringArea { /* string area for various strings */
  /* eventually, should allow efficient access to namespaces... :-) */
  b4302StringListType *pStrElem; /* element names */
  b4302StringSListType *psslStrAttr; /* attribute names, 
					with lists of values if not too long */
} b4302StringAreaType;

typedef struct b4302Attr {
  char *pszName;
  /* char *pszPrefix; */
  /* char *pszNSID; */
  char *pszValue;
  struct b4302Attr *pNext;
  /* struct b4302Attr *pLeft; */ 
  struct b4302Elem *pMyElem;
} b4302AttrType;

typedef struct b4302Elem {
  char *pszName;
  /* char *pszPrefix; */
  /* char *pszNSID; */
  b4302AttrType *pAttrFirst;
  struct b4302Elem *pLeft;
  struct b4302Elem *pNext;
  struct b4302Elem *pDown;
  /* struct b4302Elem *pLastSon; *//* handled in code - it's not so bad... */
  struct b4302Elem *pParent; /* redundant but useful for effective traversal */
  /* struct b4302Doc *pDoc; *//* this document */
} b4302ElemType;

typedef struct b4302StackItem {
  struct b4302StackItem *pNext; /* next item down the stack */
  struct b4302Elem *pLastElem; /* last element on level: for efficiency only 
				  if NULL: first son of pParent should 
				  be inserted next and kept in *this* item! */
  int iNonBlankText; /* set if some non-blank text on this level */
} b4302StackItemType;

typedef struct b4302IOOptions {
  char szCode[200]; /* unused so far; normalized charset string */
                    /* could a charset string be longer than 200?? */
  int fParseErrorsOK; /* 1: return tree even if input has parsing
			 error somewhere in the middle;
			 0: destroy tree on error, return nonzero code
			 or NULL for (b4302DocType *) functions */
  int fNoBlanks; /* remove removable blanks from document */
                 /* NB: blanks removed from before and after main doc always */
                 /* values: 
		    0 preserve blanks everywhere inside doc 
		    1 xmllint comaptible (not impl yet): remove blank 
		      text nodes until nonblank text node found, 
		      then do keep all blank text nodes (at every level) 
                    2 aggressive: remove all blank text nodes */
  int iIndent; /* output indentation; works only for file output, not
		  for ...ToString... functions
		  values: 
		  -1 no indent
		  >= 0: difference in indentation of tree levels */
  /* more to come I guess... */
} b4302IOOptionsType;

typedef struct b4302Doc {
  b4302IOOptionsType *poCurr; /* parsing options, could be NULL if none */
  char *pszName; /* not yet sure what this will be used for... */
  FILE *fDocIn; /* optional, used for reading */
  FILE *fDocOut; /* optional, used for writing */
  b4302ElemType *pRoot; /* root of the document tree */
  b4302ElemType *peFirstFragment; /* linked list of fragments bound
				     to the same document; points to 
				     virtual root of such fragment;
				     virtual root uses pNext for linking */
                                  /* is peNULL if only main docuemnt exists */
  int cElements; /* number of elements in tree, inc. text nodes */
  int cAttrs; /* number of attributes */
  /* next items needed for parsing/printing only: */ 
  char *pszCurrentFile; /* file being worked on, in or out */
  b4302StackItemType *pParsingStack; /* sem.: pNext is next down the stack,
					pDown is ptr to open elem node (might
					be NULL - if first on level),
					pParent is parent of pDown */
  b4302StringAreaType *sStringArea; /* stringarea (first item) for document */
  int fDocumentIncompleteAfterParse; /* set by parser if premature end of
					input or other error but an option 
					allowed it
					not to consider it a fatal error */
  char *szXML; /***/ /* for first line */
  char *szDocument; /***/ /* for <!DOCUMENT line */
} b4302DocType;

/* getting around the document: */
#define b4302NodeIsElement(pElem) (*(pElem->pszName) != '<' && *(pElem->pszName) != PSEUDOOTHER && *(pElem->pszName) != '>')
#define b4302NodeIsText(pElem) (*(pElem->pszName) == '<')
#define b4302NodeIsComment(pElem) (*(pElem->pszName) == PSEUDOOTHER \
              && !strcmp(pElem->pszName,COMMENTPSEUDOTAGNAME))
#define b4302NodeIsCDATA(pElem) (*(pElem->pszName) == PSEUDOOTHER \
              && !strcmp(pElem->pszName,CDATAPSEUDOTAGNAME))
#define b4302NodeIsVirtualRoot(pElem) (*(pElem->pszName) == '>')

inline int b4302ElementNameIsInvalid(char *pszName);

inline b4302ElemType *b4302DocGetRootElement(b4302DocType *pDoc);

inline b4302ElemType *b4302GetNextElement(b4302ElemType *pElem);

inline b4302ElemType *b4302GetPrevElement(b4302ElemType *pElem);

inline b4302ElemType *b4302GetFirstChild(b4302ElemType *pElem);

inline b4302ElemType *b4302GetLastChild(b4302ElemType *pElem);

inline b4302ElemType *b4302GetParent(b4302ElemType *pElem);

inline b4302AttrType *b4302GetFirstAttr(b4302ElemType *pElem);

inline b4302AttrType *b4302GetNextAttr(b4302AttrType *pAttr);

int b4302BindFragment(b4302DocType *pd, b4302ElemType *peVR);

int b4302InsertBefore(b4302ElemType *peParent, 
		      b4302ElemType *peNew, 
		      b4302ElemType *peRef);

int b4302InsertAfter(b4302ElemType *peParent, 
		      b4302ElemType *peNew, 
		      b4302ElemType *peRef);

b4302ElemType *b4302RemoveChild(b4302ElemType *peParent, 
				b4302ElemType *peChildToRemove);

/* NB: ReplaceChild if not successful might still have removed
   old node before returning error status! */
b4302ElemType *b4302ReplaceChild(b4302ElemType *peParent, 
				 b4302ElemType *peNew,
				 b4302ElemType *peOld);

b4302ElemType *b4302CreateElement(b4302DocType *pd, char *pszElemName);

b4302ElemType *b4302CreateTextNode(b4302DocType *pd, char *pszString);
b4302ElemType *b4302CreateCommentNode(b4302DocType *pd, char *pszString);

int b4302DestroyNode(b4302ElemType *peNode);

int b4302DestroySubtree(b4302ElemType *peSub);

int b4302DestroyDoc(b4302DocType *pd);

b4302AttrType * b4302GetAttributeNode(b4302ElemType *pElem, 
				      const char *pszAttrName);
inline char * b4302GetAttribute(b4302ElemType *pElem, const char *pszAttrName);

inline b4302AttrType * b4302DocSetAttribute(b4302DocType *pd,
					    b4302ElemType *pElem, 
					    const char *pszAttrName, 
					    const char *pszAttrValue);
inline b4302AttrType * b4302SetAttribute(b4302ElemType *pElem, 
					 const char *pszAttrName, 
					 const char *pszAttrValue);

int b4302DocRemoveAttribute(b4302DocType *pd,
			    b4302ElemType *pElem, 
			    const char *pszAttrName);
int b4302RemoveAttribute(b4302ElemType *pElem, 
			 const char *pszAttrName);

char *b4302GetData(b4302ElemType *pElem);
char *b4302SetData(b4302ElemType *pElem, const char *pszString);
b4302ElemType *b4302DocAppendText(b4302DocType *pd,
				  b4302ElemType *peParent, 
				  const char *pszString);
b4302ElemType *b4302AppendText(b4302ElemType *pElemParent, 
			       const char *pszString);
b4302ElemType *b4302DocAppendTextChild(b4302DocType *pd, 
				       b4302ElemType *peParent, 
				       const char *pszElemName,
				       const char *pszString);
b4302ElemType *b4302AppendTextChild(b4302ElemType *pElemParent, 
				    const char *pszElemName,
				    const char *pszString);

char *b4302StringOfElement(b4302ElemType *pe); /* xpath notion of 'string' */

/* search nodes: */
b4302ElemType *b4302FindFirstChild(b4302ElemType *peParent, char *pszStr);
b4302ElemType *b4302FindNextSibling(b4302ElemType *peNode, char *pszStr);
b4302ElemType *b4302FindPrevSibling(b4302ElemType *peNode, char *pszStr);
b4302ElemType *b4302FindAncestor(b4302ElemType *peChild, char *pszStr);

int /* reading a document from a file; */
b4302ReadFile(b4302DocType **ppDoc, char *pszFileName, b4302IOOptionsType *po);

int /* write a document to a file; */
b4302WriteFile(b4302DocType *pDoc, char *pszFileName);

/* more convenient reading document from a file; */
b4302DocType *b4302ParseFile(char *pszFileName, b4302IOOptionsType *po);

/* parsing from a string: */
b4302DocType *b4302ParseString(char *pszStringXML, b4302IOOptionsType *po);
b4302ElemType *b4302DocParseString(b4302DocType *pd, char *pszStringXML);

/* I/O options for document parsing/printing: */
b4302IOOptionsType *b4302NewOptions(void);
void b4302DeleteOptions(b4302IOOptionsType *);
void b4302SetOptionCode(b4302IOOptionsType *po, char *pszCode);
void b4302SetOptionParseErrorsOK(b4302IOOptionsType *po, int fFlag);
void b4302SetOptionNoBlanks(b4302IOOptionsType *po, int fFlag);
void b4302SetOptionIndent(b4302IOOptionsType *po, int i);

char *b4302ToString(b4302ElemType *pe);

char *b4302GetElementName(b4302ElemType *pElem);

#define B4302LIBH
#endif



